//Jason Kim
//Lua Scripting Integration with C++ Example.  Functions,Classes,Globals/Variables!
//Change values in the test.lua script file to see scripting in action!

#include <iostream>

using namespace std;

//Get Lua
extern "C"
{
	#include "lua.h"
	#include "lualib.h"
	#include "lauxlib.h"
}

#include "luna.h" //a template class available at http://lua-users.org used for binding c++ classes to Lua!

//The Lua Interpreter/Virtual Machine
lua_State* L;


//C++ function that Lua script can call 
//apparently lua fuctions in C cannot be void however, must return something!
int MyFunction(lua_State *L)
{
	int numargs = lua_gettop(L);
	cout << "Lua called MyFunction, with " << numargs << " arguments" << endl;

	for( int n=1; n <= numargs; ++n)
	{
		cout << "Arguement #" << n << ": " << lua_tostring(L,n) << endl;
	}
	return 0;
}

//Conversely, C++ can call functions defined in Lua
int LuaFunctionAdd(int x, int y)
{
	int sum;

	lua_getglobal(L,"add");//get the function name
	lua_pushnumber(L,x);//first arguement
	lua_pushnumber(L,y);//second arguement
	lua_call(L,2,1);//call function with 2 arguements and return 1 arguement
	sum = (int)lua_tointeger(L, -1);//get the result
	lua_pop(L, 1);
	return sum;

}

//Lua and Classes using Luna
class Player
{
public:
	static const char className[];
	static Luna<Player>::RegType methods[];

	//Constructor/Destructor
	Player(lua_State *L) { PlayerID = luaL_checknumber(L,1); Level = luaL_checknumber(L,2); cout << "Player " << PlayerID << " at Level " << Level << " Created" << endl; Health = 9999; Strength = 999; Agility = 999; Intelligence = 999;}
	~Player() { cout << "Player " << PlayerID << " at Level " << Level << " Deleted" << endl;} 
	
	//Methods
	//Note that the Script cannot directly access the variables in private class but must have a function to it
	int SetPlayerValues(lua_State *L) { Health = luaL_checknumber(L,1); Strength = luaL_checknumber(L,2); Agility = luaL_checknumber(L,3); Intelligence = luaL_checknumber(L,4); return 1;}	
	int PlayerFunction(lua_State *L){ cout << "Player " << PlayerID << " Called PlayerFunction" << endl; return 0;}
	int GetHealth(lua_State *L) { lua_pushnumber(L, Health); return 1; }
	int GetStrength(lua_State *L) { lua_pushnumber(L, Strength); return 1;}
	int GetAgility(lua_State *L) { lua_pushnumber(L, Agility); return 1;}
	int GetIntelligence(lua_State *L) { lua_pushnumber(L, Intelligence); return 1;}
	int GetPlayerID(lua_State *L) { lua_pushnumber(L, PlayerID); return 1;}
	int GetLevel(lua_State *L){ lua_pushnumber(L,Level); return 1;}
private:
	lua_Number Health;
	lua_Number Strength;
	lua_Number Agility;
	lua_Number Intelligence;
	lua_Number PlayerID;
	lua_Number Level;
};

//register class methods in Lua
const char Player::className[] = "Player";
#define method(class, name) {#name, &class::name}
Luna<Player>::RegType Player::methods[] = 
{
	method(Player, SetPlayerValues),
	method(Player, PlayerFunction),
	method(Player, GetHealth),
	method(Player, GetStrength),
	method(Player, GetAgility),
	method(Player, GetIntelligence),
	method(Player, GetPlayerID),
	method(Player, GetLevel),
	{0,0}
};

//our main!
void main()
{
	
	//init lua Virtual Machine
	L = lua_open();

	//open Lua Base Libraries
	luaL_openlibs(L);

	//register player class
	Luna<Player>::Register(L);

	//register MyFunction with Lua script
	lua_register(L,"MyFunction", MyFunction);

	//run script file
	luaL_dofile(L, "test.lua");

	cout << endl;
	cout << "----------------------";
	cout << endl;


	//C++ calls a function defined in script
	int sum;
	//call the script function in C++
	sum = LuaFunctionAdd( 2,2 );
	cout << "The sum of 2 + 2 is: " << sum << endl;
	cout << "This function was called in C++ and defined in lua" << endl;


	cout << endl;
	cout << "----------------------";
	cout << endl;

	//effecting c++ variables in lua such as gravity in a game
	float gravity = 2.1f;
	cout << "Gravity value in C++ is: " << gravity << " m/s^2" << endl;
	
	//get the global "gravity" from lua to change our gravity to 9.8m/s^2 the gravity of the earth
	lua_getglobal(L,"gravity");
	gravity = (float)lua_tonumber(L,-1);

	cout << "Gravity value adjusted by Lua is: " << gravity << " m/s^2" << endl;

	cout << endl;
	cout << "----------------------";
	cout << endl;

	
	//cleanup lua
	lua_close(L);


	cout << endl;
	cout << "Hit Any Key To Exit" << endl;
	system("pause");
}